from  collective.smsauthenticator.testing import COLLECTIVE_SMSAUTHENTICATOR_FUNCTIONAL_TESTING
from plone.testing import layered
import robotsuite
import unittest


def test_suite():
    suite = unittest.TestSuite()
    # suite.addTests([
    #     layered(robotsuite.RobotTestSuite("robot_test.txt"),
    #             layer=COLLECTIVE_SMSAUTHENTICATOR_FUNCTIONAL_TESTING)
    # ])
    return suite


from uuid import uuid4

from zope.i18nmessageid import MessageFactory

from collective.smsauthenticator.helpers import get_app_settings
from collective.smsauthenticator.pas_plugin import SMSAuthenticatorPlugin

_ = MessageFactory('collective.smsauthenticator')

PAS_TITLE = 'SMS Authenticator plugin (collective.smsauthenticator)'
PAS_ID = 'sms_auth'

def _add_plugin(pas, pluginid=PAS_ID):
    """
    Install and activate collective.smsauthenticator PAS plugin
    """
    installed = pas.objectIds()
    if pluginid in installed:
        return PAS_TITLE + " already installed."
    plugin = SMSAuthenticatorPlugin(pluginid, title=PAS_TITLE)
    pas._setObject(pluginid, plugin)
    plugin = pas[plugin.getId()] # get plugin acquisition wrapped!
    for info in pas.plugins.listPluginTypeInfo():
        interface = info['interface']
        if not interface.providedBy(plugin):
            continue
        pas.plugins.activatePlugin(interface, plugin.getId())
        pas.plugins.movePluginsDown(
            interface,
            [x[0] for x in pas.plugins.listPlugins(interface)[:-1]],
        )

def _setup_secret_key(portal):
    """
    Generate secret key
    """
    portal.portal_setup.runImportStepFromProfile(
        'profile-collective.smsauthenticator:default',
        'plone.app.registry'
        )

    settings = get_app_settings()
    if not settings.ska_secret_key:
        settings.ska_secret_key = unicode(uuid4())

def setupVarious(context):
    """
    @param context: Products.GenericSetup.context.DirectoryImportContext instance
    """

    # We check from our GenericSetup context whether we are running
    # add-on installation for your product or any other proudct
    if context.readDataFile('collective.smsauthenticator.marker.txt') is None:
        # Not your add-on
        return

    portal = context.getSite()

    _setup_secret_key(portal)

    pas = portal.acl_users
    _add_plugin(pas)




"""
User mobile number setup. Here user himself sets the two-step verification.
"""

import logging

from zope.i18nmessageid import MessageFactory

from z3c.form import button, field

from plone.directives import form
from plone import api
from plone.z3cform.layout import wrap_form

from Products.statusmessages.interfaces import IStatusMessage
from zope.schema import TextLine

from collective.smsauthenticator.helpers import generate_code, send_mobile_number_setup_confirmation_code_sms

logger = logging.getLogger('collective.smsauthenticator')

_ = MessageFactory('collective.smsauthenticator')


class ISetupMobileNumberForm(form.Schema):
    """
    Interface for the SMS Authenticator mobile number setup form.
    """
    mobile_number = TextLine(
        title=_(u"Mobile number"),
        description=_(u"Enter your mobile number to be used for two-step verification."),
        required=True
    )


class SetupMobileNumberForm(form.SchemaForm):
    """
    Form for the SMS Authenticator setup.
    """
    fields = field.Fields(ISetupMobileNumberForm)
    ignoreContext = True
    schema = ISetupMobileNumberForm
    label = _("Setup mobile number for two-step verification")
    description = _(u"To setup two-step verification you need to enter your mobile phone number"
                    u"to which you would be receiving SMS messages with login codes.")

    @button.buttonAndHandler(_('Verify'))
    def handleSubmit(self, action):
        if bool(api.user.is_anonymous()) is True:
            self.request.response.setStatus(401, _('Forbidden for anonymous'), True)
            return False

        data, errors = self.extractData()
        if errors:
            return False

        mobile_number = data.get('mobile_number', '')
        reason = None

        if mobile_number:
            try:
                # Set the ``enable_two_step_verification`` to True
                user = api.user.get_current()

                # Send a code to confirm the mobile number
                mobile_number_confirmation_code = generate_code(user)

                # Send the SMS
                send_mobile_number_setup_confirmation_code_sms(
                    mobile_number = mobile_number,
                    code = mobile_number_confirmation_code
                    )

                user.setMemberProperties(
                    mapping = {
                        'mobile_number': mobile_number,
                        'mobile_number_reset_code': mobile_number_confirmation_code,
                    }
                    )

                IStatusMessage(self.request).addStatusMessage(
                    _("A confirmation SMS message has been sent to the mobile number specified."),
                    'info'
                    )
                redirect_url = "{0}/@@setup-two-step-verification".format(self.context.absolute_url())
            except Exception as e:
                reason = _(str(e))
        else:
            reason = _("Invalid mobile number.")

        if reason is not None:
            IStatusMessage(self.request).addStatusMessage(_("Setup failed! {0}".format(reason)), 'error')
            redirect_url = "{0}/@@setup-mobile-number".format(self.context.absolute_url())

        self.request.response.redirect(redirect_url)

    def updateFields(self, *args, **kwargs):
        """
        Bar code image is applied here.
        """
        if bool(api.user.is_anonymous()) is False:

            # Something happens here...

            return super(SetupMobileNumberForm, self).updateFields(*args, **kwargs)


# View for the ``SetupMobileNumberForm``.
SetupMobileNumberFormView = wrap_form(SetupMobileNumberForm)



# -*- extra stuff goes here -*-

from zope.i18nmessageid import MessageFactory

from Products.PluggableAuthService.PluggableAuthService import registerMultiPlugin

from collective.smsauthenticator.pas_plugin import (
    SMSAuthenticatorPlugin, addSMSAuthenticatorPlugin, manage_addSMSAuthenticatorPluginForm
    )

_ = MessageFactory('collective.smsauthenticator')

def initialize(context):
    """
    Initializer called when used as a Zope 2 product.
    """
    registerMultiPlugin(SMSAuthenticatorPlugin.meta_type) # Add to PAS menu
    context.registerClass(
        SMSAuthenticatorPlugin,
        constructors = (manage_addSMSAuthenticatorPluginForm, addSMSAuthenticatorPlugin),
        visibility = None
        )



import logging

from zope.interface import Interface, implements

from plone.app.users.browser.personalpreferences import UserDataPanelAdapter
from collective.smsauthenticator.helpers import extract_next_url_from_referer

logger = logging.getLogger(__file__)

class EnhancedUserDataPanelAdapter(UserDataPanelAdapter):
    """
    Adapter for `collective.smsauthenticator.userdataschema.IEnhancedUserDataSchema`.
    """
    # ****************************************************
    # ******* ``enable_two_step_verification`` *******
    # ****************************************************
    def get_enable_two_step_verification(self):
        return self.context.getProperty('enable_two_step_verification', '')

    def set_enable_two_step_verification(self, value):
        return # Read only
        #return self.context.setMemberProperties({'enable_two_step_verification': value})

    enable_two_step_verification = property(
        get_enable_two_step_verification,
        set_enable_two_step_verification
        )

    # ****************************************************
    # ***************** ``mobile_number`` ****************
    # ****************************************************
    def get_mobile_number(self):
        return self.context.getProperty('mobile_number', '')

    def set_mobile_number(self, value):
        return # Read only

    mobile_number = property(
        get_mobile_number,
        set_mobile_number
        )

    # ****************************************************
    # ********** ``two_step_verification_secret`` ****
    # ****************************************************
    def get_two_step_verification_secret(self):
        return self.context.getProperty('two_step_verification_secret', '')

    def set_two_step_verification_secret(self, value):
        return # Read only

    two_step_verification_secret = property(
        get_two_step_verification_secret,
        set_two_step_verification_secret
        )

    # ****************************************************
    # ********** ``mobile_number_reset_token`` ***********
    # ****************************************************
    def get_mobile_number_reset_token(self):
        return self.context.getProperty('mobile_number_reset_token', '')

    def set_mobile_number_reset_token(self, value):
        return # Read only

    mobile_number_reset_token = property(
        get_mobile_number_reset_token,
        set_mobile_number_reset_token
        )

    # ****************************************************
    # *********** ``mobile_number_reset_code`` ***********
    # ****************************************************
    def get_mobile_number_reset_code(self):
        return self.context.getProperty('mobile_number_reset_code', '')

    def set_mobile_number_reset_code(self, value):
        return # Read only

    mobile_number_reset_code = property(
        get_mobile_number_reset_code,
        set_mobile_number_reset_code
        )

    # ****************************************************
    # ********* ``authentication_token_valid_until`` *****
    # ****************************************************
    #def get_authentication_token_valid_until(self):
    #    return self.context.getProperty('authentication_token_valid_until', '')
    #
    #def set_authentication_token_valid_until(self, value):
    #    return # Read only
    #
    #authentication_token_valid_until = property(
    #    get_authentication_token_valid_until,
    #    set_authentication_token_valid_until
    #    )

    # ****************************************************
    # ******** ``mobile_number_authentication_code`` *****
    # ****************************************************
    def get_mobile_number_authentication_code(self):
        return self.context.getProperty('mobile_number_authentication_code', '')

    def set_mobile_number_authentication_code(self, value):
        return # Read only

    mobile_number_authentication_code = property(
        get_mobile_number_authentication_code,
        set_mobile_number_authentication_code
        )


class ICameFrom(Interface):
    """
    Interface for getting the ``came_from`` URL.
    """


class CameFromAdapter(object):
    """
    Came from handling.

    Plone `came_from` field had to be taken out of the login form, so that users always get the
    token validation screen, prior to being redirected to page they came from. The came_from
    is instead extracted from referer and handled in such a way, that Plone functionality stays
    intact.

    In cases your existing package smuggles with `came_from` (for example, you want users first
    to accept terms and conditions prior redirection), you would likely need to define
    a new adapter and make appropriate changes to the ``getCameFrom`` method.

    :example:
    >>> from zope.interface import implements
    >>> from plone import api
    >>> from collective.smsauthenticator.helpers import extract_next_url_from_referer
    >>> from collective.smsauthenticator.adapter import ICameFrom
    >>> 
    >>> class CameFromAdapter(object):
    >>>     implements(ICameFrom)
    >>> 
    >>>     def __init__(self, request):
    >>>         self.request = request
    >>> 
    >>>     def getCameFrom(self):
    >>>         real_referrer = extract_next_url_from_referer(self.request)
    >>>         portal = api.portal.get()
    >>>         if not real_referrer:
    >>>             real_referrer = portal.absolute_url()
    >>>         referrer = "{0}/tac-form/?came_from={1}".format(portal.portal_url(), real_referrer)
    >>>         return referrer
    """
    implements(ICameFrom)

    def __init__(self, request):
        """
        :param request ZPublisher.HTTPRequest:
        """
        self.request = request

    def getCameFrom(self):
        """
        Extracts the ``came_from`` value from the referrer (uses global request).

        :return string:
        """
        return extract_next_url_from_referer(self.request)



from zope.i18nmessageid import MessageFactory

from plone import api

from Products.Five import BrowserView
from Products.statusmessages.interfaces import IStatusMessage

from collective.smsauthenticator.helpers import disable_two_step_verification_for_users

_ = MessageFactory('collective.smsauthenticator')

class DisableTwoStepVerificationForAllUsers(BrowserView):
    """
    Disable the two-step verification for all users.
    """
    def __init__(self, context, request):
        self.context = context
        self.request = request

    def index(self):
        """
        Disble the two-step verification for the user and redirect back to the `@@sms-authenticator-settings`.
        """
        users = api.user.get_users()
        disable_two_step_verification_for_users(users)

        IStatusMessage(self.request).addStatusMessage(
            _("You have successfully disabled the two-step verification for all users."),
            'info'
            )
        redirect_url = "{0}/@@sms-authenticator-settings".format(self.context.absolute_url())
        self.request.response.redirect(redirect_url)



"""
Reset bar-code. This is the place where the new number is actually set (upon confirmation). First of
all upon landing on the page (GET request) the validity of the URL is checked, since it's signed
and has an expiration limit of 10 minutes. The phone number to be added to users' profile is
taken from the GET request as well, but no worries - it's signed too, thus, in case of data tumpering,
the entire signature would become invalid. Note, that signature used for resetting the mobile number is
written to the users' profile on the moment request is sent. No unauthorised reproduction is possible,
since we match it to the one in users' profile.

If all things mentioned above are in good order, user would see a form where he can fill in the
verification code received by SMS. That code had been stored in users' profile too and is simply
matched the one filled in.

If all checks pass well, the mobile number is definitely set in users' profile.
"""
import logging

from zope.i18nmessageid import MessageFactory

from z3c.form import button, field

from plone.directives import form
from plone import api
from plone.z3cform.layout import wrap_form

from Products.statusmessages.interfaces import IStatusMessage
from zope.schema import TextLine

from collective.smsauthenticator.helpers import validate_mobile_number_reset_code, validate_user_data

logger = logging.getLogger('collective.smsauthenticator')

_ = MessageFactory('collective.smsauthenticator')


class IResetMobileNumberForm(form.Schema):
    """
    Interface for the SMS Authenticator Reset Mobile Number form.
    """
    token = TextLine(
        title=_(u"Enter the verification code sent to you by SMS to activate two-step verification."),
        description=_(u"You have (or will shortly) received an SMS message with verification code "
                      u'to reset your mobile number. Enter the code below.'),
        required=True
    )


class ResetMobileNumberForm(form.SchemaForm):
    """
    Form for the resetting the mobile number.

    What happens here is:

    - Signed user data is validated. If valid, the user is fetched.
    - Token (`signature` param) is matched to the one obtained from user records. If matched, the
      mobile number is reset (saved in the users' profile).
    """
    fields = field.Fields(IResetMobileNumberForm)
    ignoreContext = True
    schema = IResetMobileNumberForm
    label = _("Reset your two-step verification mobile number")
    description = _(u"To reset the two-step verification mobile number you need to enter the code "
                    u"you have received on the mobile number provided.")

    def action(self):
        return "{0}?{1}".format(
            self.request.getURL(),
            self.request.get('QUERY_STRING', '')
            )

    @button.buttonAndHandler(_('Verify'))
    def handleSubmit(self, action):
        data, errors = self.extractData()
        if errors:
            return False

        # Mobile number specified as a new phone number
        mobile_number = self.request.get('mobile_number', '')

        # Token entered by user from his mobile device sent to him by SMS
        token = data.get('token', '')

        # Signature token generated for resetting the bar code.
        signature_token = self.request.get('signature', '')
        valid_until = self.request.get('valid_until', '')

        username = self.request.get('auth_user', '')
        user = api.user.get(username=username)

        if not user:
            reason = _("User not found {0}.".format(username))
            IStatusMessage(self.request).addStatusMessage(
                _("Resetting of the mobile number failed! {0}".format(reason)),
                'error'
                )
            return

        # Validating the SMSAuthenticator app token
        valid_token = validate_mobile_number_reset_code(token, user=user)

        reason = None
        if valid_token:
            try:
                # Checking if token generated for resetting the bar code image is equal
                # to the one taken from current request.
                mobile_number_reset_token = user.getProperty('mobile_number_reset_token')
                if mobile_number_reset_token != signature_token:
                    reason = _("Invalid mobile number reset token.")
                    IStatusMessage(self.request).addStatusMessage(
                        _("Resetting of the mobile number failed! {0}".format(reason)),
                        'error'
                        )
                    return

                user.setMemberProperties(
                    mapping = {
                        'enable_two_step_verification': True,
                        'mobile_number': mobile_number,
                        'mobile_number_reset_token': '',
                        'mobile_number_reset_code': ''
                        }
                    )

                IStatusMessage(self.request).addStatusMessage(
                    _("Two-step verification mobile number is successfully reset for your account."),
                    'info'
                    )
                redirect_url = "{0}".format(self.context.absolute_url())
                self.request.response.redirect(redirect_url)
            except Exception as e:
                reason = _(str(e))
        else:
            reason = _("Invalid token or token expired.")

        if reason is not None:
            IStatusMessage(self.request).addStatusMessage(_("Setup failed! {0}".format(reason)), 'error')

    def updateFields(self, *args, **kwargs):
        """
        Here happens the following:

        - Signed user data is validated. If valid, the user is fetched.
        - Token (`signature` param) is matched to the one obtained from user records. If matched, the
          mobile number is reset (and saved in the users' profile).
        """
        username = self.request.get('auth_user', '')
        token = self.request.get('signature', '')
        user = api.user.get(username=username)

        # If valid user
        if user:
            # Getting the users' mobile number reset token saved in his profile.
            mobile_number_reset_token = user.getProperty('mobile_number_reset_token')

            # Validate the user data
            user_data_validation_result = validate_user_data(request=self.request, user=user)

            # If all goes well, save the number in profile (overwrite_secret=True).
            if not user_data_validation_result.result:
                IStatusMessage(self.request).addStatusMessage(
                    ' '.join(user_data_validation_result.reason),
                    'error'
                    )
            if mobile_number_reset_token != token:
                IStatusMessage(self.request).addStatusMessage(
                    _("Invalid mobile number reset token"),
                    'error'
                    )

        return super(ResetMobileNumberForm, self).updateFields(*args, **kwargs)


# View for the ``ResetMobileNumberForm``.
ResetMobileNumberFormView = wrap_form(ResetMobileNumberForm)



from plone.app.testing import PloneSandboxLayer
from plone.app.testing import applyProfile
from plone.app.testing import PLONE_FIXTURE
from plone.app.testing import IntegrationTesting
from plone.app.testing import FunctionalTesting

from plone.testing import z2

from zope.configuration import xmlconfig


class CollectivesmsauthenticatorLayer(PloneSandboxLayer):

    defaultBases = (PLONE_FIXTURE,)

    def setUpZope(self, app, configurationContext):
        # Load ZCML
        import collective.smsauthenticator
        xmlconfig.file(
            'configure.zcml',
            collective.smsauthenticator,
            context=configurationContext
        )

        # Install products that use an old-style initialize() function
        z2.installProduct(app, 'collective.smsauthenticator')

#    def tearDownZope(self, app):
#        # Uninstall products installed above
#        z2.uninstallProduct(app, 'collective.smsauthenticator')


COLLECTIVE_SMSAUTHENTICATOR_FIXTURE = CollectivesmsauthenticatorLayer()
COLLECTIVE_SMSAUTHENTICATOR_INTEGRATION_TESTING = IntegrationTesting(
    bases=(COLLECTIVE_SMSAUTHENTICATOR_FIXTURE,),
    name="CollectivesmsauthenticatorLayer:Integration"
)
COLLECTIVE_SMSAUTHENTICATOR_FUNCTIONAL_TESTING = FunctionalTesting(
    bases=(COLLECTIVE_SMSAUTHENTICATOR_FIXTURE, z2.ZSERVER_FIXTURE),
    name="CollectivesmsauthenticatorLayer:Functional"
)



from zope.i18nmessageid import MessageFactory

from plone import api

from Products.Five import BrowserView
from Products.statusmessages.interfaces import IStatusMessage

_ = MessageFactory('collective.smsauthenticator')

class DisableTwoStepVerification(BrowserView):
    """
    Disabling the two-step verification.
    """
    def __init__(self, context, request):
        self.context = context
        self.request = request

    def disable(self):
        """
        Disable the two-step verification for the user and redirect back to the `@@personal-information`.
        """
        if bool(api.user.is_anonymous()) is True:
            self.request.response.setStatus(401, _('Forbidden for anonymous'), True)
            return None

        user = api.user.get_current()
        user.setMemberProperties(
            mapping = {
                'enable_two_step_verification': False,
                'two_step_verification_secret': '',
                'mobile_number_reset_token': '',
                'mobile_number_reset_code': '',
                #'authentication_token_valid_until': '',
                'mobile_number_authentication_code': '',
            }
            )

        IStatusMessage(self.request).addStatusMessage(
            _("You have successfully disabled the two-step verification for your account."),
            'info'
            )
        # TODO: redirect user to where he actually came from?
        redirect_url = "{0}/@@personal-information".format(self.context.absolute_url())
        self.request.response.redirect(redirect_url)



"""
The idea of this PAS plugin is quite simple. It should check the user profile for the user being logged in
and if user has enabled two-step verification for his account (``enable_two_step_verification`` is set to
True), then redirect him further to a another page, where he would enter his SMS Authenticator token, after
successful validation of which the user would be definitely logged in.

If user has not enabled the two-step verification for his account (``enable_two_step_verification``
is set to False), then do nothing so that Plone continues logging in the user normal way.
"""
import logging

from Globals import InitializeClass
from AccessControl.SecurityInfo import ClassSecurityInfo

from zope.i18nmessageid import MessageFactory

from plone import api

from Products.PluggableAuthService.plugins.BasePlugin import BasePlugin
from Products.PluggableAuthService.utils import classImplements
from Products.PluggableAuthService.interfaces.plugins import IAuthenticationPlugin
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from Products.CMFCore.utils import getToolByName
from Products.statusmessages.interfaces import IStatusMessage

from collective.smsauthenticator.adapter import ICameFrom
from collective.smsauthenticator.helpers import (
    sign_user_data, is_whitelisted_client, generate_code, send_login_code_sms
    )

logger = logging.getLogger("collective.smsauthenticator")

_ = MessageFactory('collective.smsauthenticator')

manage_addSMSAuthenticatorPluginForm = PageTemplateFile(
    './www/add_sms_authenticator_form',
    globals(),
    __name__ = 'manage_addSMSAuthenticatorPluginForm'
    )

def addSMSAuthenticatorPlugin(self, id, title='', REQUEST=None):
    """
    Add a SMS Authenticator PAS Plugin to Plone PAS
    """
    o = SMSAuthenticatorPlugin(id, title)
    self._setObject(o.getId(), o)

    if REQUEST is not None:
        REQUEST['RESPONSE'].redirect(
            '{0}/manage_main?manage_tabs_message=SMS+Authentiactor+PAS+Plugin+added.'.format(self.absolute_url())
            )

class SMSAuthenticatorPlugin(BasePlugin):
    """
    SMS Authenticator PAS Plugin
    """
    meta_type = 'Collective SMS Authenticator PAS'
    security = ClassSecurityInfo()

    def __init__(self, id, title=None):
        self._setId(id)
        self.title = title

    def authenticateCredentials(self, credentials):
        """
        Place to actually validate the user credentials specified and return a tuple (login, login)
        on success or (None, None) on failure.

        If we find one and two-step verification is not enabled for the account, we consider
        the authentication passed and log the user in. If two-step verification has been enabled
        for the account, the first step of authentication is considered to be passed and we go to
        the next page (having the user and pass remembered), where we check for the token generated by
        the token generator (SMS Authenticator). If the token is valid too, we log the user in.
        """
        login = credentials['login']
        password = credentials['password']

        if not login:
            return (None, None)

        user = api.user.get(username=login)

        #logger.debug("Found user: {0}".format(user.getProperty('username')))

        two_step_verification_enabled = user.getProperty('enable_two_step_verification')
        #logger.debug("Two-step verification enabled: {0}".format(two_step_verification_enabled))

        if two_step_verification_enabled:
            # First see, if the password is correct.
            mt = getToolByName(self, 'portal_membership')
            plone_auth_user = None
            try:
                # This is where the password check is made.
                plone_auth_user = mt.authenticate(login, password, self.REQUEST)

            except Exception as e:
                plone_auth_user = None

            if not plone_auth_user:
                # Password check failed.
                return None

            if is_whitelisted_client():
                return (None, None)

            # Setting the data in the session doesn't seem to work. That's why we use the `ska` package.
            # The secret key would be then a combination of username, secret stored in users' profile
            # and the browser version.
            request = self.REQUEST
            response = request['RESPONSE']
            response.setCookie('__ac', '', path='/')

            # Redirect to token thing...
            signed_url = sign_user_data(request=request, user=user, url='@@sms-authenticator-token')

            came_from_adapter = ICameFrom(request)
            # Appending possible `came_from`, but give it another name.
            came_from = came_from_adapter.getCameFrom()
            if came_from:
                signed_url = '{0}&next_url={1}'.format(signed_url, came_from)

            # ****************************************
            # Generate the login code and send the SMS
            uuu = api.user.get(username=login)
            mobile_number = uuu.getProperty('mobile_number')
            mobile_number_authentication_code = uuu.getProperty('mobile_number_authentication_code')

            # If user doesn't have yet a mobile numer specified, redirect him to the reset mobile
            # number view.
            if not mobile_number:
                IStatusMessage(request).addStatusMessage(
                    _("Two-step verification is enabled for your account, but you haven't specified "
                      "a mobile number yet! In order to be able to log in, you have to go through mobile "
                      "number recovery procedure first."), 'warning'
                    )
                response.redirect('@@request-mobile-number-reset', lock=1)
                return None

            # If code was not yet sent, generate a code and send an SMS.
            if not mobile_number_authentication_code:
                mobile_number_authentication_code = generate_code(uuu)

                # Send the SMS
                sms_sent = send_login_code_sms(
                    mobile_number = mobile_number,
                    code = mobile_number_authentication_code
                    )

                if sms_sent:
                    # Save the `signature` value to the `mobile_number_reset_token`.
                    uuu.setMemberProperties(
                        mapping = {
                            'mobile_number_authentication_code': mobile_number_authentication_code,
                        }
                        )
                #else:
                #    IStatusMessage(request).addStatusMessage(
                #        _("An error occured while sending the SMS to the number given."), 'error'
                #        )
                #    response.redirect('login', lock=1)
                #    return None

                    #logger.debug("mobile number: {0}".format(mobile_number))

            # Redirecting user to authentication code validation page.
            response.redirect(signed_url, lock=1)

            return None

        if credentials.get('extractor') != self.getId():
            return (None, None)

        return (None, None)


classImplements(SMSAuthenticatorPlugin, IAuthenticationPlugin)
InitializeClass(SMSAuthenticatorPlugin)



import logging

from zope.component import getUtility

from zope.i18nmessageid import MessageFactory
from zope.interface import Interface
from zope.schema import TextLine, Bool, Text, Int

from plone.registry.interfaces import IRegistry
from plone import api
from plone.app.registry.browser import controlpanel
from plone.autoform.form import AutoExtensibleForm
from plone.directives.form import fieldset

from z3c.form import form, button

from Products.statusmessages.interfaces import IStatusMessage

logger = logging.getLogger("collective.smsauthenticator")

_ = MessageFactory('collective.smsauthenticator')


class ISMSAuthenticatorSettings(Interface):
    """
    Global SMS Authenticator settings.
    """
    # Local settings
    globally_enabled = Bool(
        title = _("Globally enabled"),
        description = _("If checked, globally enables the two-step verification for all users; "
                        "otherwise - each user configures it himself. Note, that unchecking the "
                        "checkbox does not disable the two-step verification for all users."),
        required = False,
        default = True,
        )
    ip_addresses_whitelist = Text(
        title = _("White-listed IP addresses"),
        description = _("Two-step verification will be ommit for users that log in from white "
                        "listed addresses."),
        required = False,
        default = u'',
        )

    fieldset(
        'main',
        label=_("Main"),
        fields=['globally_enabled', 'ip_addresses_whitelist',]
        )

    # Twilio settings
    twilio_number = TextLine(
        title = _("Twilio number"),
        description = _("Enter your Twilio (phone) number."),
        required = True,
        default = u'',
        )

    twilio_account_sid = TextLine(
        title = _("Twilio AccountSID"),
        description = _("Enter your Twilio AccountSID."),
        required = True,
        default = u'',
        )

    twilio_auth_token = TextLine(
        title = _("Twilio AuthToken"),
        description = _("Enter your Twilio AuthToken."),
        required = True,
        default = u'',
        )

    fieldset(
        'twilio',
        label=_("Twilio"),
        fields=['twilio_number', 'twilio_account_sid', 'twilio_auth_token',]
        )

    # Security settings
    ska_secret_key = TextLine(
        title = _("Secret Key"),
        description = _("Enter your secret key for the site here. When choosing a secret key, "
                        "think of it as some sort of a password."),
        required = False,
        default = u'',
        )

    ska_token_lifetime = Int(
        title = _("Token lifetime"),
        description = _("Authentication token lifetime in seconds."),
        required = False,
        default = 300, # 15 minutes
        )

    fieldset(
        'security',
        label=_("Security"),
        fields=['ska_secret_key', 'ska_token_lifetime',]
        )


class SMSAuthenticatorSettingsEditForm(AutoExtensibleForm, form.EditForm):
    """
    Control panel form.
    """
    control_panel_view = "plone_control_panel"
    schema_prefix = None
    schema = ISMSAuthenticatorSettings
    label = _("SMS Authenticator")
    description = _(u"""SMS Authenticator configuration""")

    def updateFields(self):
        super(SMSAuthenticatorSettingsEditForm, self).updateFields()

    def updateWidgets(self):
        super(SMSAuthenticatorSettingsEditForm, self).updateWidgets()

    def getContent(self):
        return getUtility(IRegistry).forInterface(self.schema, prefix=self.schema_prefix)

    def updateActions(self):
        super(SMSAuthenticatorSettingsEditForm, self).updateActions()
        self.actions['save'].addClass("context")
        self.actions['cancel'].addClass("standalone")

    def render(self, *args, **kwargs):
        res = super(SMSAuthenticatorSettingsEditForm, self).render(*args, **kwargs)
        additional_template = self.context.restrictedTraverse('control_panel_extra')
        additional = additional_template(
            enable_url = '{0}/{1}'.format(self.context.absolute_url(), '@@sms-authenticator-enable-for-all-users'),
            enable_text = _("Enable two-step verification for all users"),
            disable_url = '{0}/{1}'.format(self.context.absolute_url(), '@@sms-authenticator-disable-for-all-users'),
            disable_text = _("Disable two-step verification for all users"),
            charset = 'utf-8',
            )
        return res + additional

    @button.buttonAndHandler(_(u"Save"), name='save')
    def handleSave(self, action):
        """
        Update properties of all users.
        """
        from collective.smsauthenticator.helpers import (
            enable_two_step_verification_for_users, disable_two_step_verification_for_users
            )
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return

        globally_enabled = data.get('globally_enabled', None)

        if globally_enabled is True:
            # Enable for all users
            users = api.user.get_users()
            enable_two_step_verification_for_users(users)
            logger.debug('Enabled')
        elif globally_enabled is False:
            # Disable for all users
            users = api.user.get_users()
            #disable_two_step_verification_for_users(users)
            logger.debug('Disabled')

        changes = self.applyChanges(data)
        IStatusMessage(self.request).addStatusMessage(_(u"Changes saved."), "info")
        self.request.response.redirect("%s/%s" % (self.context.absolute_url(), self.control_panel_view))

    @button.buttonAndHandler(_(u"Cancel"), name='cancel')
    def handleCancel(self, action):
        IStatusMessage(self.request).addStatusMessage(_(u"Edit cancelled."), "info")
        self.request.response.redirect("%s/%s" % (self.context.absolute_url(), self.control_panel_view))


class SMSAuthenticatorSettingsControlPanel(controlpanel.ControlPanelFormWrapper):
    """
    Control panel.
    """
    form = SMSAuthenticatorSettingsEditForm



from Products.CMFCore.utils import getToolByName
import unittest2 as unittest
from plone.testing.z2 import Browser
from plone.app.testing import quickInstallProduct
from plone.app.testing import SITE_OWNER_NAME, SITE_OWNER_PASSWORD, TEST_USER_NAME, TEST_USER_PASSWORD
from plone import api

from collective.smsauthenticator.testing import \
    COLLECTIVE_SMSAUTHENTICATOR_INTEGRATION_TESTING
from collective.smsauthenticator.tests.base import BaseTest


class TestGeneric(unittest.TestCase, BaseTest):

    layer = COLLECTIVE_SMSAUTHENTICATOR_INTEGRATION_TESTING

    def setUp(self):
        self.app = self.layer['app']
        self.portal = self.layer['portal']
        self.qi_tool = getToolByName(self.portal, 'portal_quickinstaller')
        self.portal_url = api.portal.get().absolute_url()
        self._install()

    def test_(self):
        """
        """
        


from plone.testing.z2 import Browser
from plone.app.testing import SITE_OWNER_NAME, SITE_OWNER_PASSWORD

class BaseTest(object):

    def _install(self):

        browser = Browser(self.app)

        # Login as site owner
        browser.open('{0}/login_form'.format(self.portal.absolute_url()))
        browser.getControl(name='__ac_name').value = SITE_OWNER_NAME
        browser.getControl(name='__ac_password').value = SITE_OWNER_PASSWORD
        browser.getControl(name='submit').click()

        # We must uninstall and install the package, it seems generic setup profile
        # is not applied coorectly by plone.app.testing in this testing layer.
        browser.open('{0}/prefs_install_products_form'.format(self.portal.absolute_url()))

        form = browser.getForm(index=1)
        self.assertEqual(
            form.action, '{0}/portal_quickinstaller/installProducts'.format(self.portal.absolute_url()),
            u'Install form not found')
        products_list = form.getControl(name='products:list')
        if "collective.smsauthenticator" in products_list.options:
            products_list.value = (u"collective.smsauthenticator",)
            form.getControl(label='Activate').click()
        browser.open(self.portal.absolute_url() + '/logout')

    def _get_browser(self):
        browser = Browser(self.app)
        browser.handleErrors = False
        return browser

    def _login_browser(self, browser, user, passwd):
        browser.open(self.portal_url + '/login_form')
        browser.getControl(name='__ac_name').value = user
        browser.getControl(name='__ac_password').value = passwd
        browser.getControl(name='submit').click()


"""
User setup. Here user himself sets the two-step verification.
"""

import logging

from zope.i18nmessageid import MessageFactory

from z3c.form import button, field

from plone.directives import form
from plone import api
from plone.z3cform.layout import wrap_form

from Products.statusmessages.interfaces import IStatusMessage
from zope.schema import TextLine

from collective.smsauthenticator.helpers import validate_mobile_number_reset_code, get_or_create_secret

logger = logging.getLogger('collective.smsauthenticator')

_ = MessageFactory('collective.smsauthenticator')


class ISetupTwoStepVerificationForm(form.Schema):
    """
    Interface for the SMS Authenticator setup form.
    """
    token = TextLine(
        title=_(u"Enter the verification code to activate two-step verification "),
        description=_(u"You should have received an SMS message to the mobile number specified. "
                      u"Please, enter the code below."),
        required=True
    )


class SetupTwoStepVerificationForm(form.SchemaForm):
    """
    Form for the SMS Authenticator setup.
    """
    fields = field.Fields(ISetupTwoStepVerificationForm)
    ignoreContext = True
    schema = ISetupTwoStepVerificationForm
    label = _("Setup two-step verification")
    description = _(u"Complete the two-step verification setup by confirming your mobile number")

    @button.buttonAndHandler(_('Verify'))
    def handleSubmit(self, action):
        if bool(api.user.is_anonymous()) is True:
            self.request.response.setStatus(401, _('Forbidden for anonymous'), True)
            return False

        data, errors = self.extractData()
        if errors:
            return False

        token = data.get('token', '')

        user = api.user.get_current()

        # Validating the SMSAuthenticator app token
        valid_token = validate_mobile_number_reset_code(token, user=user)

        reason = None
        if valid_token:
            try:
                # Set the ``enable_two_step_verification`` to True
                two_step_verification_secret = get_or_create_secret(user)
                user.setMemberProperties(
                    mapping = {
                        'enable_two_step_verification': True,
                        'two_step_verification_secret': two_step_verification_secret,
                        'mobile_number_reset_code': '',
                    }
                    )

                IStatusMessage(self.request).addStatusMessage(
                    _("Two-step verification is successfully enabled for your account."),
                    'info'
                    )
                redirect_url = "{0}/@@personal-information".format(self.context.absolute_url())
            except Exception as e:
                reason = _(str(e))
        else:
            reason = _("Invalid token or token expired.")

        if reason is not None:
            IStatusMessage(self.request).addStatusMessage(_("Setup failed! {0}".format(reason)), 'error')
            redirect_url = "{0}/@@setup-two-step-verification".format(self.context.absolute_url())

        self.request.response.redirect(redirect_url)

    def updateFields(self, *args, **kwargs):
        """
        Check if user has a mobile number. If not, redirect him to the page to set it up.
        """
        if bool(api.user.is_anonymous()) is False:
            user = api.user.get_current()
            mobile_number = user.getProperty('mobile_number')
            if not mobile_number:
                redirect_url = "{0}/@@setup-mobile-number".format(self.context.absolute_url())
                self.request.response.redirect(redirect_url)

            return super(SetupTwoStepVerificationForm, self).updateFields(*args, **kwargs)

# View for the ``SetupForm``.
SetupTwoStepVerificationFormView = wrap_form(SetupTwoStepVerificationForm)



"""
Request the mobile number reset. What actually happens here, is that user provides his username
and a mobile number. After that an confirmation email is sent to users' email address (taken
from the profile). Also, an SMS message with code to reset the mobile number is immediately
sent to the phone number given. No other information is specified in the SMS. Once user checks
his email and follows the link given, he lands on the page on which he is supposed to fill in the
verification code received by SMS. Upon successful validation, the mobile number is reset.
"""
import logging
from smtplib import SMTPRecipientsRefused

from zope.i18nmessageid import MessageFactory
from zope.schema import TextLine
from z3c.form import button, field

from plone.directives import form
from plone import api
from plone.z3cform.layout import wrap_form

from Products.statusmessages.interfaces import IStatusMessage
from Products.CMFCore.utils import getToolByName

from ska import Signature, RequestHelper

from collective.smsauthenticator.helpers import (
    get_ska_secret_key, generate_code, send_mobile_number_reset_confirmation_code_sms,
    get_ska_token_lifetime
    )

logger = logging.getLogger('collective.smsauthenticator')

_ = MessageFactory('collective.smsauthenticator')


class IRequestMobileNumberResetForm(form.Schema):
    """
    Interface for the request to reset the SMS Authenticator mobile number form.
    """
    username = TextLine(
        title=_(u'Username'),
        description=_(u'Enter your username for verification. The link code to reset the '
                      u'mobile number would be sent to your email.'),
        required=True
    )
    mobile_number = TextLine(
        title=_(u'Mobile number'),
        description=_(u"Enter your mobile phone number. A verification code would be sent "
                      u"to that number by SMS shortly."),
        required=True
    )


class RequestMobileNumberResetForm(form.SchemaForm):
    """
    Form for request to reset to the SMS Authenticator mobile number form.
    """
    fields = field.Fields(IRequestMobileNumberResetForm)
    ignoreContext = True
    schema = IRequestMobileNumberResetForm
    label = _("Request to reset the mobile number")
    description = _(u"Enter your username for verification. The link code to reset the "
                    u"mobile number would be sent to your email.")

    @button.buttonAndHandler(_('Submit'))
    def handleSubmit(self, action):
        data, errors = self.extractData()
        if errors:
            return False

        username = data.get('username', '')
        mobile_number = data.get('mobile_number', '')

        user = api.user.get(username=username)

        reason = None
        if user:
            try:
                # Here we need to generate a token which is valid for let's say, 2 hours
                # using which it should be possible to reset the mobile number. The `signature`
                # generated should be saved in the user profile `mobile_number_reset_token`.
                ska_secret_key = get_ska_secret_key(request=self.request, user=user)

                # We also need to generate another token (no security, just a random string)
                # to sent to users' mobile number.
                mobile_number_reset_code = generate_code(user)

                token_lifetime = get_ska_token_lifetime()

                signature = Signature.generate_signature(
                    auth_user = username,
                    secret_key = ska_secret_key,
                    lifetime = token_lifetime,
                    extra = {'mobile_number': mobile_number}
                    )

                request_helper = RequestHelper(
                    signature_param = 'signature',
                    auth_user_param = 'auth_user',
                    valid_until_param = 'valid_until'
                    )

                signed_url = request_helper.signature_to_url(
                    signature = signature,
                    endpoint_url = '{0}/{1}'.format(self.context.absolute_url(), '@@reset-mobile-number')
                )

                # Send the SMS
                sms_sent = send_mobile_number_reset_confirmation_code_sms(
                    mobile_number = mobile_number,
                    code = mobile_number_reset_code
                    )

                if not sms_sent:
                    IStatusMessage(self.request).addStatusMessage(
                        _("An error occured while sending the SMS to the number specified."),
                        'info'
                        )
                    redirect_url = "{0}".format(self.context.absolute_url())
                    self.request.response.redirect(redirect_url)
                    return

                # Save the `signature` value to the `mobile_number_reset_token`.
                user.setMemberProperties(
                    mapping = {
                        'mobile_number_reset_token': str(signature),
                        'mobile_number_reset_code': mobile_number_reset_code,
                        'mobile_number_authentication_code': '',
                        }
                    )

                # Now we need to send an email to user with URL in and a small explanations.
                try:
                    host = getToolByName(self, 'MailHost')

                    mail_text_template = self.context.restrictedTraverse('request_mobile_number_reset_email')
                    mail_text = mail_text_template(
                        member = user,
                        mobile_number_reset_url = signed_url,
                        charset = 'utf-8'
                        )
                    mail_text = mail_text.format(mobile_number_reset_url=signed_url)

                    host.send(
                        mail_text,
                        immediate = True,
                        msg_type = 'text/html'
                        )
                except SMTPRecipientsRefused as e:
                    raise SMTPRecipientsRefused('Recipient address rejected by server')

                IStatusMessage(self.request).addStatusMessage(
                    _("An email with instructions on resetting your mobile number is sent successfully."),
                    'info'
                    )
                redirect_url = "{0}".format(self.context.absolute_url())
                self.request.response.redirect(redirect_url)
            except ValueError as e:
                reason = _(str(e))
        else:
            reason = _("Invalid username.")

        if reason is not None:
            IStatusMessage(self.request).addStatusMessage(
                _("Request for mobile number reset is failed! {0}".format(reason)),
                'error'
                )

    def updateFields(self, *args, **kwargs):
        """
        """
        request = self.request
        response = request['RESPONSE']
        response.setCookie('__ac', '', path='/')

        return super(RequestMobileNumberResetForm, self).updateFields(*args, **kwargs)


# View for the ``RequestMobileNumberResetForm``.
RequestMobileNumberResetFormView = wrap_form(RequestMobileNumberResetForm)



from setuptools import setup, find_packages
import os

version = '0.1'

try:
    readme = open('README.rst').read()
    readme = readme.replace('.. image:: _static', '.. figure:: https://github.com/collective/collective.smsauthenticator/raw/master/docs/_static')
except:
    readme = ''

try:
    changelog = open('CHANGES.txt').read()
except:
    changelog = ''

long_description = (
    readme
    + '\n' +
    #'Contributors\n'
    #'============\n'
    #+ '\n' +
    #open('CONTRIBUTORS.txt').read()
    #+ '\n' +
    changelog
+ '\n')

setup(
    name = 'collective.smsauthenticator',
    version = version,
    description = "Two-step verification in Plone 4 using login codes sent by SMS.",
    long_description = long_description,
    # Get more strings from
    # http://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers = [
        "Environment :: Web Environment",
        "Framework :: Plone",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords = 'sms authenticator, two-step verification, multi-factor authentication, two-factor authentication',
    author = 'Goldmund, Wyldebeast & Wunderliebe',
    author_email = 'info@gw20e.com',
    url = 'https://github.com/collective/collective.smsauthenticator',
    license = 'GPL 2.0',
    packages = find_packages('src'),
    package_dir = {'': 'src'},
    namespace_packages = ['collective', ],
    include_package_data = True,
    zip_safe = False,
    install_requires = [
        'setuptools',
        # -*- Extra requirements: -*-
        'plone.api==1.1.0',
        'plone.directives.form>=1.1',
        'ska==1.4.2',
        'rebus==0.2',
        'twilio>=3.6.0'
    ],
    extras_require = {'test': ['plone.app.testing', 'robotsuite']},
    entry_points = """
        # -*- Entry points: -*-
        [z3c.autoinclude.plugin]
        target = plone
    """,
)



import logging

from plone import api

from zope.component import adapter
from zope.schema import Bool, TextLine
from zope.i18nmessageid import MessageFactory
from zope.interface import implements

from plone.app.users.userdataschema import IUserDataSchema, IUserDataSchemaProvider
from plone.app.users.browser.personalpreferences import UserDataPanel

from Products.PluggableAuthService.interfaces.authservice import IBasicUser
from Products.PluggableAuthService.interfaces.events import IPrincipalCreatedEvent

logger = logging.getLogger("collective.smsauthenticator")

_ = MessageFactory('collective.smsauthenticator')

class CustomizedUserDataPanel(UserDataPanel):
    """
    Customise the user form shown in personal-preferences.
    """
    def __init__(self, context, request):
        super(CustomizedUserDataPanel, self).__init__(context, request)

        # Removing certain fields from form
        self.form_fields = self.form_fields.omit(
            'enable_two_step_verification',
            'mobile_number',
            'two_step_verification_secret',
            'mobile_number_reset_token',
            'mobile_number_reset_code',
            'mobile_number_authentication_code',
            #'authentication_token_valid_until',
            )


class UserDataSchemaProvider(object):
    implements(IUserDataSchemaProvider)

    def getSchema(self):
        """
        """
        return IEnhancedUserDataSchema


class IEnhancedUserDataSchema(IUserDataSchema):
    """
    Extended user profile.

    :property bool enable_two_step_verification: Indicates, whether the two-step verification is
                                                     enabled for the user.
    :property string two_step_verification_secret: Secret key of the user (unique per user).
                                                       Automatically generated.
    :property string mobile_number_reset_token: Token to reset users' mobile number. Automatically
                                                generated.
    :property string mobile_number_reset_code: Code to reset users' mobile number (sent by SMS).
                                               Automatically generated.
    :property string mobile_number_authentication_code: Latest code sent by SMS to users' mobile
                                                        number for authentication. Automatically
                                                        generated.
    """
    enable_two_step_verification = Bool(
        title=_('Enable two-step verification.'),
        description=_("""Enable/disable the two-step verification. Click """
                      """<a href=\"@@setup-mobile-number\">here</a> to set it up or """
                      """<a href=\"@@disable-two-step-verification\">here</a> to disable it."""
            ),
        required=False
        )

    mobile_number = TextLine(
        title = _('Mobile number'),
        description = _("Users' mobile number"),
        required = False,
    )

    two_step_verification_secret = TextLine(
        title = _('Secret key'),
        description = _('Automatically generated'),
        required = False,
    )

    mobile_number_reset_token = TextLine(
        title = _('Token to reset the mobile number'),
        description = _('Automatically generated'),
        required = False,
    )

    mobile_number_reset_code = TextLine(
        title = _('Code to reset the mobile number (sent by SMS)'),
        description = _('Automatically generated'),
        required = False,
    )

    mobile_number_authentication_code = TextLine(
        title = _('Last authentication code'),
        description = _('Automatically generated each time SMS message is sent'),
        required = False,
    )

    #authentication_token_valid_until = TextLine(
    #    title = _('Last token valid until'),
    #    description = _('Automatically generated each time SMS message is sent'),
    #    required = False,
    #)


@adapter(IBasicUser, IPrincipalCreatedEvent)
def userCreatedHandler(principal, event):
    """
    Fired upon creation of each user. If app setting ``globally_enabled`` is set to True,
    two-step verification would be automatically enabled for the registered users (in that
    case they would have to go through the mobile number recovery procedure.

    The ``principal`` value is seems to be a user object, although it does not have
    the ``setMemberProperties`` method defined (that's why we obtain the user
    using `plone.api`, 'cause that one has it).
    """
    from collective.smsauthenticator.helpers import (
        is_two_step_verification_globally_enabled, get_or_create_secret
        )
    user = api.user.get(username=principal.getId())
    if is_two_step_verification_globally_enabled():
        get_or_create_secret(user)
        user.setMemberProperties(mapping={'enable_two_step_verification': True,})

    logger.debug(user.getProperty('enable_two_step_verification'))
    logger.debug(user.getProperty('two_step_verification_secret'))



from zope.i18nmessageid import MessageFactory

from plone import api

from Products.Five import BrowserView

from collective.smsauthenticator.helpers import (
    is_two_step_verification_globally_enabled, has_enabled_two_step_verification
    )

_ = MessageFactory('collective.smsauthenticator')

class SettingsHelper(BrowserView):
    """
    Helper view for accessing some conditions from portal actions (actions.xml).
    """
    def __init__(self, context, request):
        self.context = context
        self.request = request

    def is_two_step_verification_globally_enabled(self):
        """
        Disable the two-step verification for the user and redirect back to the `@@personal-information`.
        """
        return is_two_step_verification_globally_enabled()

    def show_enable_two_step_verification_link(self):
        """
        Indicates whether the enable two-step verification link should be shown.

        The following conditions shall be met for True to be returned:

        - User hasn't enabled the two-step verification for his account.
        - In app settings, the globally enable two-step verification is set to False.

        :return bool:
        """
        user = api.user.get_current()
        return not is_two_step_verification_globally_enabled() \
               and (has_enabled_two_step_verification(user) is False)

    def show_disable_two_step_verification_link(self):
        """
        Indicates whether the disable two-step verification link should be shown.

        The following conditions shall be met for True to be returned:

        - User hasn enabled the two-step verification for his account.
        - In app settings, the globally enable two-step verification is set to False.

        :return bool:
        """
        user = api.user.get_current()
        return not is_two_step_verification_globally_enabled() and has_enabled_two_step_verification(user)



"""
The helper module contains various methods for api security and for downloading files
"""
from hashlib import sha1
from uuid import uuid4
from urllib import unquote, quote
from urlparse import urlparse
import logging
from random import randint

from zope.component import getUtility
from zope.globalrequest import getRequest
from zope.i18nmessageid import MessageFactory

from twilio.rest import TwilioRestClient

from plone.registry.interfaces import IRegistry
from plone import api

from ska import sign_url, validate_signed_request_data
import rebus

from collective.smsauthenticator.browser.controlpanel import ISMSAuthenticatorSettings

_ = MessageFactory('collective.smsauthenticator')

logger = logging.getLogger("collective.smsauthenticator")

# ******************************************

def get_app_settings():
    """
    Gets the SMS Authenticator settings.
    """
    registry = getUtility(IRegistry)
    settings = registry.forInterface(ISMSAuthenticatorSettings)
    return settings

def get_user(username):
    """
    Get user by username given and return member object.
    """
    return api.user.get(username=username)

def get_username(user=None):
    """
    Gets the username of the user.

    :param user: If given, used to extract the user. Otherwise, ``plone.api.user.get_current`` is used.
    :return string:
    """
    if user is None:
        user = api.user.get_current()
    if user:
        return user.getUserName()

def get_base_url(request=None):
    """
    Gets domain name (with HTTP).

    :param ZPublisher.HTTPRequest request:
    :return string:
    """
    if request is None:
        request = getRequest()

    parsed_uri = urlparse(request.base)
    return "{0}://{1}/".format(parsed_uri.scheme, parsed_uri.netloc)

def get_domain_name(request=None):
    """
    Gets domain name (without HTTP).

    :param ZPublisher.HTTPRequest request:
    :return string:
    """
    if request is None:
        request = getRequest()

    parsed_uri = urlparse(request.base)
    return parsed_uri.netloc

def generate_secret(user):
    """
    Generates secret for the user.

    :param Products.PlonePAS.tools.memberdata user:
    """
    secret = rebus.b32encode(str(uuid4()))
    user.setMemberProperties(mapping={'two_step_verification_secret': secret,})
    return secret

def get_secret(user=None, hashed=False):
    """
    Gets users' secret code. If ``hashed`` is set to True, returned hashed.

    :param Products.PlonePAS.tools.memberdata user:
    :param bool hashed: If set to True, hashed version is returned.
    :return string:
    """
    #TODO: Return hashed version if ``hashed`` is set to True.
    if user is None:
        user = api.user.get_current()
    if user:
        secret = user.getProperty('two_step_verification_secret')

        # If string returned, then it's likely a set string
        if isinstance(secret, basestring) and secret:
            return secret

def get_or_create_secret(user, overwrite=False):
    """
    Gets or creates token secret for the user given. Checks first if user given has a ``secret`` generated.
    If not, generate it for him and save it in his profile (``two_step_verification_secret``).

    :param Products.PlonePAS.tools.memberdata user: If provided, used. Otherwise ``plone.api.user.get_current``
        is used to obtain the user.
    :return string:
    """
    #TODO: Return hashed version if ``hashed`` is set to True.
    if user is None:
        user = api.user.get_current()

    if overwrite:
        return generate_secret(user)

    secret = user.getProperty('two_step_verification_secret')
    if isinstance(secret, basestring) and secret:
        return secret
    else:
        return generate_secret(user)

def generate_code(user, length=6):
    """
    Gets a random token to reset the mobile number (time based) + random char.

    :param Products.PlonePAS.tools.memberdata user:
    :param int length:
    :return string:
    """
    secret = get_or_create_secret(user)
    return sha1("{0}{1}{2}".format(str(uuid4()), str(randint(0, 9)), secret)).hexdigest()[:6]

def validate_code(code, prop, user=None):
    """
    Validates the given code by matching it with one stored in users' profile.

    :param string code:
    :param string prop:
    :param Products.PlonePAS.tools.memberdata user:
    :return bool:
    """
    if user is None:
        user = api.user.get_current()

    stored_code = user.getProperty(prop)

    return stored_code == code

def validate_mobile_number_reset_code(code, user=None):
    return validate_code(code=code, prop='mobile_number_reset_code', user=user)

def validate_mobile_number_authentication_code(code, user=None):
    return validate_code(code=code, prop='mobile_number_authentication_code', user=user)

def get_browser_hash(request=None):
    """
    Gets browser hash. Adds an extra security layer, since browser version is unlikely to be changed.

    :param ZPublisher.HTTPRequest request:
    :return string:
    """
    if request is None:
        request = getRequest()

    try:
        return sha1(request.get('HTTP_USER_AGENT')).hexdigest()
    except Exception as e:
        logger.debug(str(e))
        return ''

def get_ska_secret_key(request=None, user=None, use_browser_hash=True):
    """
    Gets the `secret_key` to be used in `ska` package.

    - Value of the ``two_step_verification_secret`` (from users' profile).
    - Browser info (hash of)
    - The SECRET set for the `ska` (use `plone.app.registry`).

    :param ZPublisher.HTTPRequest request:
    :param Products.PlonePAS.tools.memberdata user:
    :param bool use_browser_hash: If set to True, browser hash is used. Otherwise - not. Defaults to True.
    :return string:
    """
    if request is None:
        request = getRequest()

    if user is None:
        user = api.user.get_current()

    settings = get_app_settings()

    ska_secret_key = settings.ska_secret_key

    user_secret = user.getProperty('two_step_verification_secret')

    if use_browser_hash:
        browser_hash = get_browser_hash(request=request)
    else:
        browser_hash = ''

    return "{0}{1}{2}".format(user_secret, browser_hash, ska_secret_key)

def is_two_step_verification_globally_enabled():
    """
    Checks if the two-step verification is globally enabled.

    :return bool:
    """
    settings = get_app_settings()
    return settings.globally_enabled

def get_white_listed_ip_addresses():
    """
    Gets list of white-listed IP addresses.

    :return list:
    """
    settings = get_app_settings()
    ip_addresses = settings.ip_addresses_whitelist
    ip_addresses_list = ip_addresses.split('\n')
    return ip_addresses_list

def get_ska_token_lifetime(settings=None):
    """
    Gets the `ska` token lifetime (in seconds) from settings.

    :return int:
    """
    if settings is None:
        settings = get_app_settings()

    return settings.ska_token_lifetime

def sign_user_data(request=None, user=None, url='@@sms-authenticator-token'):
    """
    Signs the user data with `ska` package. The secret key is `secret_key` to be used with `ska` is a
    combination of:

    - Value of the ``two_step_verification_secret`` (from users' profile).
    - Browser info (hash of)
    - The SECRET set for the `ska` (use `plone.app.registry`).

    :param ZPublisher.HTTPRequest request:
    :param Products.PlonePAS.tools.memberdata user:
    :param string url:
    :return string:
    """
    if request is None:
        request = getRequest()

    if user is None:
        user = api.user.get_current()

    token_lifetime = get_ska_token_lifetime()

    # Make sure the secret key always exists
    get_or_create_secret(user)

    secret_key = get_ska_secret_key(request=request, user=user)
    signed_url = sign_url(
        auth_user = user.getUserId(),
        secret_key = secret_key,
        url = url,
        lifetime = token_lifetime
    )
    return signed_url

def extract_request_data_from_query_string(request_qs):
    """
    Plone seems to strip/escape some special chars (such as '+') from values and those chars are
    quite important for us. This method extracts the vars from request QUERY_STRING given and
    returns them unescaped.

    :FIXME: As stated above, for some reason Plone escapes from special chars from the values. If
    you know what the reason is and if it has some effects on security, please make the changes
    necessary.

    :param string request_qs:
    :return dict:
    """
    request_data = {}

    if not request_qs:
        return request_data

    for part in request_qs.split('&'):
        try:
            key, value = part.split('=', 1)
            request_data.update({key: unquote(value)})
        except ValueError as e:
            pass

    return request_data

def extract_request_data(request):
    """
    Plone seems to strip/escape some special chars (such as '+') from values and those chars are
    quite important for us. This method extracts the vars from request QUERY_STRING given and
    returns them unescaped.

    :FIXME: As stated above, for some reason Plone escapes from special chars from the values. If
    you know what the reason is and if it has some effects on security, please make the changes
    necessary.

    :param request ZPublisher.HTTPRequest:
    :return dict:
    """
    request_qs = request.get('QUERY_STRING')
    return extract_request_data_from_query_string(request_qs)

def extract_next_url_from_referer(request, quote_url=False):
    """
    Since we override the default Plone functionality (take out the `came_from` from the login form for a
    very strong reason), we want to make sure that for users, the "came from" functionality stays intact.
    That why, we check the referer for the `came_from` attributes and if present, redirect to that after
    successful two-step verification token validation.
    :param request ZPublisher.HTTPRequest:
    :return string: Extracted `came_from` URL.
    """
    referer = request.get('HTTP_REFERER')
    request_qs = urlparse(referer).query
    request_data = extract_request_data_from_query_string(request_qs)
    url = request_data.get('came_from', '')

    if quote_url:
        return quote(url)

    return url

def validate_user_data(request, user, use_browser_hash=True):
    """
    Validates the user data.

    :param ZPublisher.HTTPRequest request:
    :param Products.PlonePAS.tools.memberdata user:
    :return ska.SignatureValidationResult:
    """
    secret_key = get_ska_secret_key(request=request, user=user, use_browser_hash=use_browser_hash)
    validation_result = validate_signed_request_data(
        data = extract_request_data(request),
        secret_key = secret_key
        )
    return validation_result

def has_enabled_two_step_verification(user):
    """
    Checks if user has enabled the two-step verification.

    :param Products.PlonePAS.tools.memberdata user:
    :return bool:
    """
    if bool(api.user.is_anonymous()) is True:
        return None

    try:
        return user.getProperty('enable_two_step_verification', False)
    except Exception as e:
        return None

def enable_two_step_verification_for_users(users=[]):
    """
    Enable two-step verification for the list of users given.
    """
    if not users:
        users = api.user.get_users()

    for user in users:
        try:
            get_or_create_secret(user)
            if not has_enabled_two_step_verification(user):
                user.setMemberProperties(mapping={'enable_two_step_verification': True,})
        except Exception as e:
            logger.debug(str(e))

def disable_two_step_verification_for_users(users=[]):
    """
    Disable two-step verification for the list of users given.
    """
    if not users:
        users = api.user.get_users()

    for user in users:
        try:
            #get_or_create_secret(user)
            if has_enabled_two_step_verification(user):
                user.setMemberProperties(
                    mapping = {
                        'enable_two_step_verification': False,
                        'two_step_verification_secret': '',
                        'mobile_number_reset_token': '',
                        'mobile_number_reset_code': '',
                        #'authentication_token_valid_until': '',
                        'mobile_number_authentication_code': '',
                    }
                    )
        except Exception as e:
            logger.debug(str(e))

def extract_ip_address_from_request(request=None):
    """
    Extracts client's IP address from request. This is not the safest solution, since client
    may change headers.

    :param ZPublisher.HTTPRequest request:
    :return string:
    """
    if not request:
        request = getRequest()

    PRIVATE_IPS_PREFIX = ('10.', '172.', '192.', )
    ip = request.get('REMOTE_ADDR')
    x_forwarded_for = request.get('HTTP_X_FORWARDED_FOR')

    if x_forwarded_for:
        proxies = [proxy.strip() for proxy in x_forwarded_for.split(',')]

        # Remove the private ips from the beginning
        while (len(proxies) > 0 and proxies[0].startswith(PRIVATE_IPS_PREFIX)):
            proxies.pop(0)

        # Take the first ip which is not a private one (of a proxy)
        if len(proxies) > 0:
            ip = proxies[0]

    return ip

def get_ip_addresses_whitelist(request=None):
    """
    Gets IP addresses white list.

    :param ZPublisher.HTTPRequest request:
    :return list:
    """
    if not request:
        request = getRequest()

    settings = get_app_settings()

    ip_addresses_whitelist = settings.ip_addresses_whitelist

    if ip_addresses_whitelist:
        try:
            ip_addresses_whitelist = ip_addresses_whitelist.split('\n')
            ip_addresses_whitelist = [ip_address.strip() for ip_address in ip_addresses_whitelist]
        except Exception as e:
            logger.debug(str(e))
            ip_addresses_whitelist = []

    return ip_addresses_whitelist or []

def is_whitelisted_client(request=None):
    """
    Checks if client's IP address is whitelisted.

    :param ZPublisher.HTTPRequest request:
    :return bool:
    """
    ip_addresses_whitelist = get_ip_addresses_whitelist(request=request)

    ip_address = extract_ip_address_from_request(request=request)

    if ip_address in ip_addresses_whitelist:
        return True

    return False

def send_sms(mobile_number, message):
    """
    Sends an SMS to the monile number given for mobile number reset confirmation.

    :param string mobile_number:
    :param string message: Message.
    :return bool: True on success and False on failure.
    """
    settings = get_app_settings()

    sms_client = TwilioRestClient(settings.twilio_account_sid, settings.twilio_auth_token)

    try:
        sms_client.sms.messages.create(
            to = mobile_number,
            from_ = settings.twilio_number,
            body = message
            )
        return True
    except Exception as e:
        # Log in the error_log
        logger.debug(e)
        return False

def send_mobile_number_reset_confirmation_code_sms(mobile_number, code):
    """
    Sends an SMS to the monile number given for mobile number reset confirmation.

    :param string mobile_number:
    :param string code:
    :return bool: True on success and False on failure.
    """
    message = _("Use this code to confirm your mobile phone reset: {0}".format(code))
    return send_sms(mobile_number, message)

def send_login_code_sms(mobile_number, code):
    """
    Sends an SMS to the monile number given for mobile number reset confirmation.

    :param string mobile_number:
    :param string code:
    :return bool: True on success and False on failure.
    """
    message = _("Use this code to login: {0}".format(code))
    return send_sms(mobile_number, message)

def send_mobile_number_setup_confirmation_code_sms(mobile_number, code):
    """
    Sends an SMS to the monile number given for mobile number setup confirmation.

    :param string mobile_number:
    :param string code:
    :return bool: True on success and False on failure.
    """
    message = _("Use this code to confirm your mobile phone setup: {0}".format(code))
    return send_sms(mobile_number, message)



"""
Token validation. Here user is supposed to fill in the token received by SMS.
"""

import logging

from zope.schema import TextLine
from zope.i18nmessageid import MessageFactory

from z3c.form import button, field

from plone.directives import form
from plone import api
from plone.z3cform.layout import wrap_form

from Products.statusmessages.interfaces import IStatusMessage

from collective.smsauthenticator.helpers import (
    validate_mobile_number_authentication_code, validate_user_data, extract_request_data
    )

logger = logging.getLogger('collective.smsauthenticator')

DEBUG = False

_ = MessageFactory('collective.smsauthenticator')

class ITokenForm(form.Schema):
    """
    Interface for the SMS Authenticator Token validation form.
    """
    token = TextLine(
        title=_('Enter code'),
        description=_('Enter the login code sent to your mobile number.'),
        required=True)


class TokenForm(form.SchemaForm):
    """
    Form for the SMS Authenticator Token validation. Any user that has two-step verification enabled,
    uses this form upon logging in.
    """
    fields = field.Fields(ITokenForm)
    ignoreContext = True
    schema = ITokenForm
    label = _(u'Two-step verification')
    description = _(u'Confirm your login by entering the login code sent to your mobile number by SMS.')

    def action(self):
        return "{0}?{1}".format(
            self.request.getURL(),
            self.request.get('QUERY_STRING', '')
            )

    @button.buttonAndHandler(_('Verify'))
    def handleSubmit(self, action):
        """
        Here we should check couple of things:

        - If the token provided is valid.
        - If the signature contains the user data needed (username and hash made of his data are valid).

        If all is well and valid, we sudo login the user given.
        """
        data, errors = self.extractData()
        if errors:
            return False

        token = data.get('token', '')

        user = None
        username = self.request.get('auth_user', '')

        if username:
            user = api.user.get(username=username)

            # Validating the signed request data. If invalid (likely throttled with or expired), generate an
            # appropriate error message.
            user_data_validation_result = validate_user_data(request=self.request, user=user)
            if not user_data_validation_result.result:
                if 'Signature timestamp expired!' in user_data_validation_result.reason:
                    # Remove used authentication code
                    user.setMemberProperties(
                        mapping = {
                            'mobile_number_authentication_code': '',
                            }
                        )
                IStatusMessage(self.request).addStatusMessage(
                    _("Invalid data. Details: {0}".format(' '.join(user_data_validation_result.reason))), 'error'
                    )
                return

        valid_token = validate_mobile_number_authentication_code(token, user=user)

        if valid_token:
            # We should login the user here
            self.context.acl_users.session._setupSession(str(username), self.context.REQUEST.RESPONSE)

            # Remove used authentication code
            user.setMemberProperties(
                mapping = {
                    'mobile_number_authentication_code': '',
                    }
                )

            # TODO: Is there a nicer way of resolving the "@@sms_authenticator_token_form" URL?
            IStatusMessage(self.request).addStatusMessage(_("Great! You're logged in."), 'info')
            request_data = extract_request_data(self.request)
            redirect_url = request_data.get('next_url', self.context.absolute_url())
            self.request.response.redirect(redirect_url)
        else:
            IStatusMessage(self.request).addStatusMessage(_("Invalid token or token expired."), 'error')

    def updateFields(self, *args, **kwargs):
        """
        Here the following happens. Cookie set is cleared. Thus, user is no longer logged in, but only
        after his SMS Authenticator token has been validated.
        """
        logger.debug("Landed in the token hook.")

        request = self.request
        response = request['RESPONSE']
        response.setCookie('__ac', '', path='/')

        # Updating the description
        token_field = self.fields.get('token')
        if token_field:
            token_field.field.description = _(
                """Enter the login code sent to your mobile number """
                """If you have somehow lost your bar code, request a mobile number reset """
                """<a href=\"{0}/@@request-mobile-number-reset\">here</a>.""".format(self.context.absolute_url())
                )

        return super(TokenForm, self).updateFields(*args, **kwargs)


# View for the ``TokenForm``.
TokenFormView = wrap_form(TokenForm)



from Products.CMFCore.utils import getToolByName
import unittest2 as unittest
from plone.testing.z2 import Browser
from plone.app.testing import quickInstallProduct
from plone.app.testing import SITE_OWNER_NAME, SITE_OWNER_PASSWORD, TEST_USER_NAME, TEST_USER_PASSWORD
from plone import api

from collective.smsauthenticator.testing import \
    COLLECTIVE_SMSAUTHENTICATOR_INTEGRATION_TESTING
from collective.smsauthenticator.tests.base import BaseTest


class TestGeneric(unittest.TestCase, BaseTest):

    layer = COLLECTIVE_SMSAUTHENTICATOR_INTEGRATION_TESTING

    def setUp(self):
        self.app = self.layer['app']
        self.portal = self.layer['portal']
        self.qi_tool = getToolByName(self.portal, 'portal_quickinstaller')
        self.portal_url = api.portal.get().absolute_url()
        self._install()

    def test_product_is_installed(self):
        """ Validate that our products GS profile has been run and the product
            installed
        """
        pid = 'collective.smsauthenticator'
        installed = [p['id'] for p in self.qi_tool.listInstalledProducts()]
        self.assertTrue(pid in installed,
            u'package appears not to have been installed')

    def test_control_panel_view(self):
        browser = self._get_browser()
        self._login_browser(browser, SITE_OWNER_NAME, SITE_OWNER_PASSWORD)
        browser.open('{0}/@@sms-authenticator-settings'.format(self.portal_url))

        self.assertEqual(browser.headers.get('status'), '200 Ok', 'HTTP response was not 200 Ok')

    def test_user_setup_view(self):
        browser = self._get_browser()
        self._login_browser(browser, TEST_USER_NAME, TEST_USER_PASSWORD)
        browser.open('{0}/@@setup-mobile-number'.format(self.portal_url))
        self.assertEqual(browser.headers.get('status'), '200 Ok', 'HTTP response was not 200 Ok')

    def test_token_view(self):
        browser = self._get_browser()
        self._login_browser(browser, TEST_USER_NAME, TEST_USER_PASSWORD)
        browser.open('{0}/@@sms-authenticator-token'.format(self.portal_url))
        self.assertEqual(browser.headers.get('status'), '200 Ok', 'HTTP response was not 200 Ok')

    # def test_disable_view(self):
    #     browser = Browser(self.app)
    #     browser.open('{0}/@@disable-two-step-verification'.format(self.portal_url))
    #
    #     self.assertEqual(browser.headers.get('status'), '200 Ok', 'HTTP response was not 200 Ok')
    #
